from typing import Literal, NamedTuple, Sequence, TypedDict


class PowerButtonConfig(NamedTuple):
    device: str
    type: Literal["hold_emitted", "hold_isa"]
    prod_name: str
    phys: str
    hold_phys: str | None = None
    hold_grab: bool | None = None
    # ev.type, ev.code, ev.value pairs
    hold_events: Sequence[tuple[int, int, int]] | None = None


# POWER_BUTTON_NAMES = ["Power Button"]
# POWER_BUTTON_PHYS = ["LNXPWRBN", "PNP0C0C"]

PBC = PowerButtonConfig

SUPPORTED_DEVICES: Sequence[PowerButtonConfig] = [
    PBC(
        "Legion Go",
        "hold_isa",
        "83E1",
        "PNP0C0C",
        "isa0060",
        False,
        [(4, 4, 219), (1, 125, 1), (0, 0, 0)],
    )
]

# Legion go
# At device with phys=isa0060/serio0/input0
#
# event at 1702329391.831152, code 04, type 04, val 219
# event at 1702329391.831152, code 125, type 01, val 01
# event at 1702329391.831152, code 00, type 00, val 00
# event at 1702329391.834208, code 04, type 04, val 103
# event at 1702329391.834208, code 00, type 00, val 00

# event at 1702329392.152298, code 04, type 04, val 219
# event at 1702329392.152298, code 125, type 01, val 00
# event at 1702329392.152298, code 00, type 00, val 00
# event at 1702329392.156347, code 04, type 04, val 103
# event at 1702329392.156347, code 00, type 00, val 00



import select
from typing import Any, Generator, Literal, NamedTuple, Sequence

from hhd.controller import Axis, Event, Axis, Producer
import os

import logging

logger = logging.getLogger(__name__)


class ScanElement(NamedTuple):
    # Output Info
    axis: Axis | None
    # Buffer Info
    endianness: Literal["little", "big"]
    signed: bool
    bits: int
    storage_bits: int
    shift: int
    # Postprocess info
    scale: float
    offset: float
    max_val: float | None


class DeviceInfo(NamedTuple):
    dev: str
    axis: Sequence[ScanElement]
    sysfs: str


ACCEL_MAPPINGS: dict[str, tuple[Axis, float | None]] = {
    "accel_x": ("accel_z", 3),
    "accel_y": ("accel_x", 3),
    "accel_z": ("accel_y", 3),
    "timestamp": ("accel_ts", None),
}
GYRO_MAPPINGS: dict[str, tuple[Axis, float | None]] = {
    "anglvel_x": ("gyro_z", None),
    "anglvel_y": ("gyro_x", None),
    "anglvel_z": ("gyro_y", None),
    "timestamp": ("gyro_ts", None),
}


def find_sensor(sensor: str):
    IIO_BASE_DIR = "/sys/bus/iio/devices/"

    for d in os.listdir(IIO_BASE_DIR):
        if not "device" in d:
            continue

        sensor_dir = os.path.join(IIO_BASE_DIR, d)
        name_fn = os.path.join(IIO_BASE_DIR, d, "name")

        if not os.path.isfile(name_fn):
            continue

        with open(name_fn, "r") as f:
            name = f.read()

        if name.strip() == sensor:
            logger.info(f"Found device '{sensor}' at\n{sensor_dir}")
            return sensor_dir

    return None


def write_sysfs(dir: str, fn: str, val: Any):
    with open(os.path.join(dir, fn), "w") as f:
        f.write(str(val))


def read_sysfs(dir: str, fn: str):
    with open(os.path.join(dir, fn), "r") as f:
        return f.read().strip()


def prepare_dev(
    sensor_dir: str,
    type: str,
    attr: str,
    freq: int | None,
    mappings: dict[str, tuple[Axis, float | None]],
    update_trigger: bool,
) -> DeviceInfo | None:
    # Prepare device buffer
    dev = os.path.join("/dev", os.path.basename(sensor_dir))
    axis = {}
    write_sysfs(sensor_dir, "buffer/enable", 0)

    # Set sampling frequency
    if freq is not None:
        sfn = os.path.join(sensor_dir, f"in_{attr}_sampling_frequency")
        if os.path.isfile(sfn):
            write_sysfs(sensor_dir, f"in_{attr}_sampling_frequency", freq)

    # Set trigger
    if update_trigger:
        trig = None
        for s in os.listdir("/sys/bus/iio/devices/"):
            if s.startswith("trigger"):
                name = read_sysfs(os.path.join("/sys/bus/iio/devices/", s), "name")
                pref = f"{type}-dev"
                if name.startswith(pref):
                    idx = name[len(pref) :]
                    if sensor_dir.endswith(idx):
                        trig = name
                        break
        if trig:
            write_sysfs(sensor_dir, "trigger/current_trigger", trig)

    # Disable all scan elements
    for s in os.listdir(os.path.join(sensor_dir, "scan_elements")):
        if s.endswith("_en"):
            write_sysfs(sensor_dir, os.path.join("scan_elements", s), 0)

    # Selectively enable required ones and fill up buffer
    for fn in os.listdir(os.path.join(sensor_dir, "scan_elements")):
        if not fn.startswith("in_") or not fn.endswith("_en"):
            continue
        fn = fn[3:-3]

        # Prepare buffer metadata
        idx = int(read_sysfs(sensor_dir, f"scan_elements/in_{fn}_index"))
        if idx == -1:
            logger.error(
                f"Device '{dev}' element '{fn}' does not support buffered capture."
            )
            return None
        se = read_sysfs(sensor_dir, f"scan_elements/in_{fn}_type")
        endianness = "big" if se.startswith("be:") else "little"
        signed = "e:s" in se

        bits = int(se[se.index(":") + 2 : se.index(":") + 4])
        storage_bits = int(se[se.index("/") + 1 : se.index("/") + 3])
        shift = int(se[-1])

        # Prepare scan metadata
        if fn in mappings:
            ax, max_val = mappings[fn]
            write_sysfs(sensor_dir, f"scan_elements/in_{fn}_en", 1)
        else:
            ax = max_val = None

        if fn != "timestamp":
            offset = float(read_sysfs(sensor_dir, f"in_{attr}_offset"))
            scale = float(read_sysfs(sensor_dir, f"in_{attr}_scale"))
        else:
            offset = 0
            scale = 1

        axis[idx] = ScanElement(
            ax, endianness, signed, bits, storage_bits, shift, scale, offset, max_val
        )
    write_sysfs(sensor_dir, "buffer/enable", 1)

    axis_arr = tuple(axis[i] for i in sorted(axis))
    return DeviceInfo(dev, axis_arr, sensor_dir)


def close_dev(dev: DeviceInfo):
    write_sysfs(dev.sysfs, "buffer/enable", 0)


def process_scan_event(data: bytes, ofs: int, se: ScanElement):
    # TODO: Implement parsing iio fully, by adding shifting and cutoff
    d = data[ofs >> 3 : (ofs >> 3) + (se.storage_bits >> 3)]
    d = int.from_bytes(d, byteorder=se.endianness, signed=se.signed)
    # d = d >> se.shift
    # d &= (1 << se.bits) - 1
    d = d * se.scale + se.offset

    if se.max_val is not None:
        if d > 0:
            d = min(d, se.max_val)
        else:
            d = max(d, -se.max_val)
    return d


def get_size(dev: DeviceInfo):
    out = 0
    for s in dev.axis:
        if out % s.storage_bits:
            # Align bytes
            out = (out // s.storage_bits + 1) * s.storage_bits
        out += s.storage_bits
    return out >> 3


class IioReader(Producer):
    def __init__(
        self,
        type: str,
        attr: str,
        freq: int | None,
        mappings: dict[str, tuple[Axis, float | None]],
        update_trigger: bool = False,
    ) -> None:
        self.type = type
        self.attr = attr
        self.freq = freq
        self.mappings = mappings
        self.update_trigger = update_trigger
        self.fd = 0

    def open(self):
        sens_dir = find_sensor(self.type)
        if not sens_dir:
            return []

        dev = prepare_dev(
            sens_dir,
            self.type,
            self.attr,
            self.freq,
            self.mappings,
            self.update_trigger,
        )
        if not dev:
            return []

        self.buf = None
        self.prev = {}
        self.dev = dev
        self.fd = os.open(dev.dev, os.O_RDONLY)
        self.size = get_size(dev)

        return [self.fd]

    def close(self, exit: bool):
        if self.fd:
            os.close(self.fd)
            self.fd = 0
        if self.dev:
            close_dev(self.dev)
            self.dev = None
        return True

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if self.fd not in fds or not self.dev:
            return []

        data = os.read(self.fd, self.size)
        if self.buf == data:
            return []
        self.buf = data

        # Empty the buffer preventing repeated calls
        while select.select([self.fd], [], [], 0)[0]:
            data = os.read(self.fd, self.size)

        out: list[Event] = []
        ofs = 0
        for se in self.dev.axis:
            # Align bytes
            if ofs % se.storage_bits:
                ofs = (ofs // se.storage_bits + 1) * se.storage_bits

            # Grab value if required
            if se.axis:
                new_val = process_scan_event(data, ofs, se)
                if se.axis not in self.prev or self.prev[se.axis] != new_val:
                    out.append(
                        {
                            "type": "axis",
                            "code": se.axis,
                            "value": new_val,
                        }
                    )
                    self.prev[se.axis] = new_val
            ofs += se.storage_bits

        # TODO: Clean this up
        # Hide duplicate events
        # if (len(out) == 1 and out[0]['code'].endswith('_ts')):
        #     return []
        return out


class AccelImu(IioReader):
    def __init__(self, freq=None) -> None:
        super().__init__("accel_3d", "accel", freq, ACCEL_MAPPINGS)


class GyroImu(IioReader):
    def __init__(self, freq=None) -> None:
        super().__init__("gyro_3d", "anglvel", freq, GYRO_MAPPINGS)


class ForcedSampler:
    def __init__(self, devices: Sequence[str], keep_fds: bool = False) -> None:
        self.devices = devices
        self.fds = []
        self.keep_fds = keep_fds

    def open(self):
        self.fds = []
        self.paths = []
        for d in self.devices:
            f = find_sensor(d)
            if not f:
                continue
            if "accel" in d:
                p = os.path.join(f, "in_accel_x_raw")
            elif "gyro" in d:
                p = os.path.join(f, "in_anglvel_x_raw")
            else:
                continue

            self.paths.append(p)
            if self.keep_fds:
                self.fds.append(os.open(p, os.O_RDONLY | os.O_NONBLOCK))

    def sample(self):
        if self.keep_fds:
            for fd in select.select(self.fds, [], [], 0)[0]:
                os.read(fd, 20)
            for fd in select.select(self.fds, [], [], 0)[0]:
                os.lseek(fd, 0, os.SEEK_SET)
        else:
            for p in self.paths:
                with open(p, "rb") as f:
                    f.read()

    def close(self):
        for fd in self.fds:
            os.close(fd)


# class SoftwareTrigger(IioReader):
#     BEGIN_ID: int = 72
#     ATTEMPTS: int = 3

#     def __init__(self, devices: Sequence[str]) -> None:
#         self.devices = devices
#         self.old_triggers = {}

#     def open(self):
#         for id in range(
#             SoftwareTrigger.BEGIN_ID,
#             SoftwareTrigger.BEGIN_ID + SoftwareTrigger.ATTEMPTS,
#         ):
#             try:
#                 with open(
#                     "/sys/bus/iio/devices/iio_sysfs_trigger/add_trigger", "w"
#                 ) as f:
#                     f.write(str(id))
#                 break
#             except Exception as e:
#                 print(e)
#         else:
#             logger.error(f"Failed to create software trigger.")
#             return
#         self.id = id

#         self.old_triggers = {}
#         for d in self.devices:
#             s = find_sensor(d)
#             if not s:
#                 continue
#             with open(os.path.join(s, "buffer/enable"), "w") as f:
#                 f.write("0")
#             trig_fn = os.path.join(s, "trigger/current_trigger")
#             with open(trig_fn, "r") as f:
#                 self.old_triggers[trig_fn] = f.read()
#             with open(trig_fn, "w") as f:
#                 f.write(f"sysfstrig{self.id}")

#     def close(self):
#         for trig, name in self.old_triggers.items():
#             try:
#                 with open(trig, "w") as f:
#                     f.write(name)
#             except Exception:
#                 logger.error(f"Could not restore original trigger:\n{trig} to {name}")

#         try:
#             with open(
#                 "/sys/bus/iio/devices/iio_sysfs_trigger/remove_trigger", "w"
#             ) as f:
#                 f.write(str(self.id))
#         except Exception:
#             logger.error(f"Could not delete sysfs trigger with id {self.id}")


__all__ = ["IioReader", "AccelImu", "GyroImu"]



import argparse
import logging
import re
import select
import sys
import time
from typing import Sequence, cast

from hhd.controller import Button, Consumer, Event, Producer
from hhd.controller.base import Multiplexer, can_read
from hhd.controller.lib.hid import enumerate_unique
from hhd.controller.physical.evdev import GenericGamepadEvdev
from hhd.controller.physical.hidraw import GenericGamepadHidraw
from hhd.controller.physical.imu import AccelImu, GyroImu
from hhd.controller.virtual.ds5 import DualSense5Edge, TouchpadCorrectionType
from hhd.controller.virtual.uinput import UInputDevice

from .const import (
    LGO_RAW_INTERFACE_AXIS_MAP,
    LGO_RAW_INTERFACE_BTN_ESSENTIALS,
    LGO_RAW_INTERFACE_BTN_MAP,
    LGO_RAW_INTERFACE_CONFIG_MAP,
    LGO_TOUCHPAD_AXIS_MAP,
    LGO_TOUCHPAD_BUTTON_MAP,
)
from .hid import rgb_callback
from .gyro_fix import GyroFixer

ERROR_DELAY = 1

logger = logging.getLogger(__name__)

LEN_VID = 0x17EF
LEN_PIDS = {
    0x6182: "xinput",
    0x6183: "dinput",
    0x6184: "dual_dinput",
    0x6185: "fps",
}


def main(as_plugin=False):
    if not as_plugin:
        from hhd import setup_logger

        setup_logger()

    parser = argparse.ArgumentParser(
        prog="HHD: LegionGo Controller Plugin",
        description="This plugin remaps the legion go controllers to a DS5 controller and restores all functionality.",
    )
    parser.add_argument(
        "-a",
        "--d-accel",
        action="store_false",
        help="Dissable accelerometer (recommended since not used by steam, .5%% core utilisation).",
        dest="accel",
    )
    parser.add_argument(
        "-g",
        "--d-gyro",
        action="store_false",
        help="Disable gyroscope (.5%% core utilisation).",
        dest="gyro",
    )
    parser.add_argument(
        "-gf",
        "--gyro-fix",
        action="store_true",
        help="Samples the gyro to avoid needing a custom driver.",
        dest="gyro_fix",
    )
    parser.add_argument(
        "-l",
        "--swap-legion",
        action="store_true",
        help="Swaps Legion buttons with start, select.",
        dest="swap_legion",
    )
    parser.add_argument(
        "-s",
        "--share-to-qam",
        action="store_true",
        help="Maps the share button (Legion R) to Guide + A",
        dest="share_to_qam",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Prints events as they happen.",
        dest="debug",
    )
    parser.add_argument(
        "-t",
        "--touchpad",
        help='How to fit the legion go touchpad into the DS5 ("stretch", "crop_center", "crop_start", "crop_end", "contain_start", "contain_end", "contain_center")',
        default=None,
    )
    if as_plugin:
        args = parser.parse_args(sys.argv[2:])
    else:
        args = parser.parse_args()

    accel = args.accel
    gyro = args.gyro
    swap_legion = args.swap_legion
    debug = args.debug
    touchpad_mode = cast(TouchpadCorrectionType | None, args.touchpad)
    gyro_fix = args.gyro_fix
    share_to_qam = args.share_to_qam
    plugin_run(
        accel=accel,
        gyro=gyro,
        swap_legion=swap_legion,
        touchpad_mode=touchpad_mode,
        gyro_fix=gyro_fix,
        share_to_qam=share_to_qam,
        debug=debug,
    )


def plugin_run(
    accel: bool = False,
    gyro: bool = True,
    swap_legion: bool = False,
    touchpad_mode: TouchpadCorrectionType | None = "crop_end",
    gyro_fix: bool | int = True,
    share_to_qam: bool = True,
    led_support: bool = True,
    debug: bool = False,
    **_,
):
    if gyro_fix:
        gyro_fixer = GyroFixer(int(gyro_fix) if int(gyro_fix) > 10 else 100)
    else:
        gyro_fixer = None

    while True:
        try:
            controller_mode = None
            pid = None
            while not controller_mode:
                devs = enumerate_unique(LEN_VID)
                if not devs:
                    logger.error(
                        f"Legion go controllers not found, waiting {ERROR_DELAY}s."
                    )
                    time.sleep(ERROR_DELAY)
                    continue

                for d in devs:
                    if d["product_id"] in LEN_PIDS:
                        pid = d["product_id"]
                        controller_mode = LEN_PIDS[pid]
                        break
                else:
                    logger.error(
                        f"Legion go controllers not found, waiting {ERROR_DELAY}s."
                    )
                    time.sleep(ERROR_DELAY)
                    continue

            match controller_mode:
                case "xinput":
                    logger.info("Launching DS5 controller instance.")
                    if gyro_fixer:
                        gyro_fixer.open()
                    controller_loop_xinput(
                        accel=accel,
                        gyro=gyro,
                        swap_legion=swap_legion,
                        share_to_qam=share_to_qam,
                        touchpad_mode=touchpad_mode,
                        led_support=led_support,
                        debug=debug,
                    )
                case _:
                    logger.info(
                        f"Controllers in non-supported (yet) mode: {controller_mode}. Launching a shortcuts device."
                    )
                    controller_loop_rest(
                        controller_mode, pid if pid else 2, share_to_qam, debug
                    )
        except Exception as e:
            logger.error(f"Received the following error:\n{e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {ERROR_DELAY}s."
            )
            if gyro_fixer:
                gyro_fixer.close()
            time.sleep(ERROR_DELAY)
        except KeyboardInterrupt:
            if gyro_fixer:
                gyro_fixer.close()
            logger.info("Received KeyboardInterrupt, exiting...")
            return


def controller_loop_rest(mode: str, pid: int, share_to_qam: bool, debug: bool = False):
    d_raw = SelectivePassthrough(
        GenericGamepadHidraw(
            vid=[LEN_VID],
            pid=list(LEN_PIDS),
            usage_page=[0xFFA0],
            usage=[0x0001],
            report_size=64,
            axis_map=LGO_RAW_INTERFACE_AXIS_MAP,
            btn_map=LGO_RAW_INTERFACE_BTN_MAP,
            required=True,
        )
    )

    multiplexer = Multiplexer(
        dpad="analog_to_discrete",
        trigger="analog_to_discrete",
        share_to_qam=share_to_qam,
    )
    d_uinput = UInputDevice(name=f"HHD Shortcuts Device (Legion Mode: {mode})", pid=pid)

    d_shortcuts = GenericGamepadEvdev(
        vid=[LEN_VID],
        pid=list(LEN_PIDS),
        name=[re.compile(r"Legion-Controller \d-.. Keyboard")],
        required=True,
    )

    try:
        fds = []
        fds.extend(d_raw.open())
        fds.extend(d_shortcuts.open())
        fds.extend(d_uinput.open())

        while True:
            select.select(fds, [], [])
            d_shortcuts.produce(fds)
            d_uinput.produce(fds)
            evs = multiplexer.process(d_raw.produce(fds))
            if debug and evs:
                logger.info(evs)
            d_uinput.consume(evs)
    finally:
        d_shortcuts.close(True)
        d_raw.close(True)
        d_uinput.close(True)


def controller_loop_xinput(
    accel: bool = True,
    gyro: bool = True,
    swap_legion: str | bool = False,
    share_to_qam: bool = False,
    touchpad_mode: TouchpadCorrectionType | None = None,
    led_support: bool = True,
    debug: bool = False,
):
    # Output
    d_ds5 = DualSense5Edge(
        touchpad_method=touchpad_mode if touchpad_mode else "crop_end"
    )

    # Imu
    d_accel = AccelImu()
    d_gyro = GyroImu()

    # Inputs
    d_xinput = GenericGamepadEvdev(
        [0x17EF],
        [0x6182],
        ["Generic X-Box pad"],
        required=True,
    )
    d_touch = GenericGamepadEvdev(
        [0x17EF],
        [0x6182],
        ["  Legion Controller for Windows  Touchpad"],
        btn_map=LGO_TOUCHPAD_BUTTON_MAP,
        axis_map=LGO_TOUCHPAD_AXIS_MAP,
        aspect_ratio=1,
        required=True,
    )
    d_raw = SelectivePassthrough(
        GenericGamepadHidraw(
            vid=[LEN_VID],
            pid=list(LEN_PIDS),
            usage_page=[0xFFA0],
            usage=[0x0001],
            report_size=64,
            axis_map=LGO_RAW_INTERFACE_AXIS_MAP,
            btn_map=LGO_RAW_INTERFACE_BTN_MAP,
            config_map=LGO_RAW_INTERFACE_CONFIG_MAP,
            callback=rgb_callback if led_support else None,
            required=True,
        )
    )
    # Mute keyboard shortcuts, mute
    d_shortcuts = GenericGamepadEvdev(
        vid=[LEN_VID],
        pid=list(LEN_PIDS),
        name=["  Legion Controller for Windows  Keyboard"],
        # report_size=64,
        required=True,
    )

    match swap_legion:
        case True:
            swap_guide = "guide_is_select"
        case False:
            swap_guide = None
        case "l_is_start":
            swap_guide = "guide_is_start"
        case "l_is_select":
            swap_guide = "guide_is_select"
        case _:
            assert False, "Invalid value for `swap_legion`."

    multiplexer = Multiplexer(
        swap_guide=swap_guide,
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        led="main_to_sides",
        status="both_to_main",
        share_to_qam=share_to_qam,
    )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        fs = m.open()
        devs.append(m)
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        if accel:
            prepare(d_accel)
        if gyro:
            prepare(d_gyro)
        prepare(d_xinput)
        prepare(d_shortcuts)
        prepare(d_touch)
        prepare(d_raw)
        prepare(d_ds5)

        logger.info("DS5 controller instance launched, have fun!")
        while True:
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

                d_ds5.consume(evs)
                d_xinput.consume(evs)
                d_raw.consume(evs)

            # If unbounded, the total number of events per second is the sum of all
            # events generated by the producers.
            # For Legion go, that would be 100 + 100 + 500 + 30 = 730
            # Since the controllers of the legion go only update at 500hz, this is
            # wasteful.
            # By setting a target refresh rate for the report and sleeping at the
            # end, we ensure that even if multiple fds become ready close to each other
            # they are combined to the same report, limiting resource use.
            # Ideally, this rate is smaller than the report rate of the hardware controller
            # to ensure there is always a report from that ready during refresh
            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        for d in devs:
            d.close(True)


class SelectivePassthrough(Producer, Consumer):
    def __init__(
        self,
        parent,
        forward_buttons: Sequence[Button] = ("share", "mode"),
        passthrough: Sequence[Button] = list(
            next(iter(LGO_RAW_INTERFACE_BTN_ESSENTIALS.values()))
        ),
    ):
        self.parent = parent
        self.state = False

        self.forward_buttons = forward_buttons
        self.passthrough = passthrough

        self.to_disable_btn = set()
        self.to_disable_axis = set()

    def open(self) -> Sequence[int]:
        return self.parent.open()

    def close(self, exit: bool) -> bool:
        return super().close(exit)

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        evs: Sequence[Event] = self.parent.produce(fds)

        out = []
        prev_state = self.state
        for ev in evs:
            if ev["type"] == "button" and ev["code"] in self.forward_buttons:
                self.state = ev.get("value", False)

            if ev["type"] == "configuration":
                out.append(ev)
            elif ev["type"] == "button" and ev["code"] in self.passthrough:
                out.append(ev)
            elif ev["type"] == "button":
                self.to_disable_btn.add(ev["code"])
            elif ev["type"] == "axis":
                self.to_disable_axis.add(ev["code"])

        if self.state:
            # If mode is pressed, forward all events
            return evs
        elif prev_state:
            # If prev_state, meaning the user released the mode or share button
            # turn off all buttons that were pressed during it
            for btn in self.to_disable_btn:
                out.append({"type": "button", "code": btn, "value": False})
            self.to_disable_btn = set()
            for axis in self.to_disable_axis:
                out.append({"type": "axis", "code": axis, "value": 0})
            self.to_disable_axis = set()
            return out
        else:
            # Otherwise, just return the standard buttons
            return out

    def consume(self, events: Sequence[Event]):
        return self.parent.consume(events)



from .base import Consumer, Event, Producer, can_read
from .const import Axis, Button, Configuration

__all__ = [
    "Axis",
    "Button",
    "Event",
    "Configuration",
    "Consumer",
    "Producer",
    "can_read",
]



import logging
import time
from collections import defaultdict
from typing import Literal, NamedTuple, Sequence, cast

from hhd.controller import Consumer, Event, Producer
from hhd.controller.lib.common import encode_axis, set_button
from hhd.controller.lib.uhid import UhidDevice, BUS_USB, BUS_BLUETOOTH

from .const import (
    DS5_USB_AXIS_MAP,
    DS5_USB_BTN_MAP,
    DS5_BT_AXIS_MAP,
    DS5_BT_BTN_MAP,
    DS5_INPUT_REPORT_BT_OFS,
    DS5_INPUT_REPORT_USB_OFS,
    DS5_EDGE_COUNTRY,
    DS5_EDGE_DELTA_TIME_NS,
    DS5_EDGE_DESCRIPTOR_BT,
    DS5_EDGE_DESCRIPTOR_USB,
    DS5_EDGE_MAX_REPORT_FREQ,
    DS5_EDGE_MIN_REPORT_FREQ,
    DS5_EDGE_NAME,
    DS5_EDGE_PRODUCT,
    prefill_ds5_report,
    DS5_EDGE_STOCK_REPORTS,
    DS5_EDGE_TOUCH_HEIGHT,
    sign_crc32_append,
    DS5_FEATURE_CRC32_SEED,
    DS5_INPUT_CRC32_SEED,
    sign_crc32_inplace,
    DS5_EDGE_TOUCH_WIDTH,
    DS5_EDGE_VENDOR,
    DS5_EDGE_VERSION,
    patch_dpad_val,
)

REPORT_MAX_DELAY = 1 / DS5_EDGE_MIN_REPORT_FREQ
REPORT_MIN_DELAY = 1 / DS5_EDGE_MAX_REPORT_FREQ
DS5_EDGE_MIN_TIMESTAMP_INTERVAL = 1500

logger = logging.getLogger(__name__)


class TouchpadCorrection(NamedTuple):
    x_mult: float = 1
    x_ofs: float = 0
    x_clamp: tuple[float, float] = (0, 1)
    y_mult: float = 1
    y_ofs: float = 0
    y_clamp: tuple[float, float] = (0, 1)


TouchpadCorrectionType = Literal[
    "stretch",
    "crop_center",
    "crop_start",
    "crop_end",
    "contain_start",
    "contain_end",
    "contain_center",
]


def correct_touchpad(
    width: int, height: int, aspect: float, method: TouchpadCorrectionType
):
    dst = width / height
    src = aspect
    ratio = dst / src

    match method:
        case "crop_center":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=(width - new_width) / 2,
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=(height - new_height) / 2,
                )
        case "crop_start":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=0,
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=0,
                )
        case "crop_end":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=(width - new_width),
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=(height - new_height),
                )
        case "contain_center":
            if ratio > 1:
                bound = (ratio - 1) / ratio / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, y_clamp=(bound, 1 - bound)
                )
            else:
                bound = (1 - ratio) / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, x_clamp=(bound, 1 - bound)
                )
        case "contain_start":
            if ratio > 1:
                bound = (ratio - 1) / ratio
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, y_clamp=(0, 1 - bound)
                )
            else:
                bound = (1 - ratio) / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, x_clamp=(0, 1 - bound)
                )
        case "contain_end":
            if ratio > 1:
                bound = (ratio - 1) / ratio
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, y_clamp=(bound, 1)
                )
            else:
                bound = (1 - ratio) / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, x_clamp=(bound, 1)
                )
        case "stretch":
            return TouchpadCorrection(x_mult=width, y_mult=height)

    logger.error(f"Touchpad correction method '{method}' not found.")
    return TouchpadCorrection(x_mult=width, y_mult=height)


class DualSense5Edge(Producer, Consumer):
    def __init__(
        self,
        touchpad_method: TouchpadCorrectionType = "crop_end",
        use_bluetooth: bool = True,
    ) -> None:
        self.available = False
        self.report = None
        self.dev = None
        self.start = 0
        self.use_bluetooth = use_bluetooth
        self.touchpad_method: TouchpadCorrectionType = touchpad_method

        self.ofs = (
            DS5_INPUT_REPORT_BT_OFS if use_bluetooth else DS5_INPUT_REPORT_USB_OFS
        )
        self.axis_map = DS5_BT_AXIS_MAP if use_bluetooth else DS5_USB_AXIS_MAP
        self.btn_map = DS5_BT_BTN_MAP if use_bluetooth else DS5_USB_BTN_MAP

    def open(self) -> Sequence[int]:
        self.available = False
        self.report = bytearray(prefill_ds5_report(self.use_bluetooth))
        self.dev = UhidDevice(
            vid=DS5_EDGE_VENDOR,
            pid=DS5_EDGE_PRODUCT,
            bus=BUS_BLUETOOTH if self.use_bluetooth else BUS_USB,
            version=DS5_EDGE_VERSION,
            country=DS5_EDGE_COUNTRY,
            name=DS5_EDGE_NAME,
            report_descriptor=DS5_EDGE_DESCRIPTOR_BT
            if self.use_bluetooth
            else DS5_EDGE_DESCRIPTOR_USB,
        )

        self.touch_correction = correct_touchpad(
            DS5_EDGE_TOUCH_WIDTH, DS5_EDGE_TOUCH_HEIGHT, 1, self.touchpad_method
        )

        self.state: dict = defaultdict(lambda: 0)
        self.rumble = False
        self.start = time.perf_counter_ns()
        self.fd = self.dev.open()
        return [self.fd]

    def close(self, exit: bool) -> bool:
        if not exit:
            """This is a consumer, so we would deadlock if it was disabled."""
            return False

        if self.dev:
            self.dev.send_destroy()
            self.dev.close()
            self.dev = None
            self.fd = None

        return True

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if self.fd not in fds:
            return []

        # Process queued events
        out: Sequence[Event] = []
        assert self.dev
        while ev := self.dev.read_event():
            match ev["type"]:
                case "open":
                    self.available = True
                case "close":
                    self.available = False
                case "get_report":
                    if ev["rnum"] in DS5_EDGE_STOCK_REPORTS:
                        rep = DS5_EDGE_STOCK_REPORTS[ev["rnum"]]
                        if self.use_bluetooth:
                            rep = sign_crc32_append(rep, DS5_FEATURE_CRC32_SEED)
                        self.dev.send_get_report_reply(ev["id"], 0, rep)
                    else:
                        logger.warn(
                            f"DS5: Received get_report with the id (uknown): {ev['rnum']}"
                        )
                case "set_report":
                    logger.warn(
                        f"DS5: Received set_report with the id (uknown): {ev['rnum']}"
                    )
                case "output":
                    invalid = False
                    # Check report num
                    if ev["report"] != 0x01:
                        invalid = True

                    # Check report ids depending on modes
                    if not self.use_bluetooth and ev["data"][0] != 0x02:
                        invalid = True
                    if self.use_bluetooth and ev["data"][0] != 0x31:
                        invalid = True

                    if invalid:
                        logger.warn(
                            f"DS5: Received uknown output report with the following data:\n{ev['report']}: {ev['data'].hex()}"
                        )
                        continue

                    rep = ev["data"]

                    if self.use_bluetooth:
                        # skip seq_tag, tag sent by bluetooth report
                        # rest is the same
                        rep = rep[0:1] + rep[2:]

                    if rep[2] & 4:  # DS_OUTPUT_VALID_FLAG1_LIGHTBAR_CONTROL_ENABLE
                        # Led data is being set
                        led_brightness = rep[43]
                        player_leds = rep[44]
                        red = rep[45]
                        green = rep[46]
                        blue = rep[47]
                        if red == 0 and green == 0 and blue == 128:
                            # Skip playstation driver usb initialization
                            continue
                        if red == 64 and green == 0 and blue == 0:
                            # Skip playstation driver bt initialization
                            continue
                        if red == 0 and green == 0 and blue == 64:
                            # Skip SDL led initialization
                            continue
                        logger.info(f"Changing leds to RGB: {red} {green} {blue}")
                        out.append(
                            {
                                "type": "led",
                                "code": "main",
                                "mode": "solid",
                                "brightness": led_brightness / 63
                                if led_brightness
                                else 1,
                                "speed": 0,
                                "red": red,
                                "blue": blue,
                                "green": green,
                            }
                        )
                    elif (rep[39] & 2) and (rep[42] & 2):
                        # flag2 is DS_OUTPUT_VALID_FLAG2_LIGHTBAR_SETUP_CONTROL_ENABLE
                        # lightbar_setup is DS_OUTPUT_LIGHTBAR_SETUP_LIGHT_OUT
                        # FIXME: Disable for now to avoid hid_playstation messing
                        # with the leds
                        # out.append(
                        #     {
                        #         "type": "led",
                        #         "code": "main",
                        #         "mode": "disable",
                        #         "brightness": 0,
                        #         "speed": 0,
                        #         "red": 0,
                        #         "blue": 0,
                        #         "green": 0,
                        #     }
                        # )
                        pass

                    if rep[1] & 0x03 == 0x03:
                        right = rep[3]
                        left = rep[4]
                        out.append(
                            {
                                "type": "rumble",
                                "code": "main",
                                "strong_magnitude": left / 63,
                                "weak_magnitude": right / 63,
                            }
                        )
                        self.rumble = True
                    elif self.rumble:
                        self.rumble = False
                        out.append(
                            {
                                "type": "rumble",
                                "code": "main",
                                "strong_magnitude": 0,
                                "weak_magnitude": 0,
                            }
                        )
                case _:
                    logger.info(f"Received unhandled report:\n{ev}")
        return out

    def consume(self, events: Sequence[Event]):
        assert self.dev and self.report

        new_rep = bytearray(self.report)
        for ev in events:
            match ev["type"]:
                case "axis":
                    if ev["code"] in self.axis_map:
                        encode_axis(new_rep, self.axis_map[ev["code"]], ev["value"])
                    # DPAD is weird
                    match ev["code"]:
                        case "hat_x":
                            self.state["hat_x"] = ev["value"]
                            patch_dpad_val(
                                new_rep,
                                self.ofs,
                                self.state["hat_x"],
                                self.state["hat_y"],
                            )
                        case "hat_y":
                            self.state["hat_y"] = ev["value"]
                            patch_dpad_val(
                                new_rep,
                                self.ofs,
                                self.state["hat_x"],
                                self.state["hat_y"],
                            )
                        case "touchpad_x":
                            tc = self.touch_correction
                            x = int(
                                min(max(ev["value"], tc.x_clamp[0]), tc.x_clamp[1])
                                * tc.x_mult
                                + tc.x_ofs
                            )
                            new_rep[self.ofs + 33] = x & 0xFF
                            new_rep[self.ofs + 34] = (new_rep[self.ofs + 34] & 0xF0) | (
                                x >> 8
                            )
                        case "touchpad_y":
                            tc = self.touch_correction
                            y = int(
                                min(max(ev["value"], tc.y_clamp[0]), tc.y_clamp[1])
                                * tc.y_mult
                                + tc.y_ofs
                            )
                            new_rep[self.ofs + 34] = (new_rep[self.ofs + 34] & 0x0F) | (
                                (y & 0x0F) << 4
                            )
                            new_rep[self.ofs + 35] = y >> 4
                        case "gyro_ts":
                            new_rep[self.ofs + 27 : self.ofs + 31] = int(
                                ev["value"] / DS5_EDGE_DELTA_TIME_NS
                            ).to_bytes(8, byteorder="little", signed=False)[:4]
                case "button":
                    if ev["code"] in self.btn_map:
                        set_button(new_rep, self.btn_map[ev["code"]], ev["value"])
                case "configuration":
                    match ev["code"]:
                        case "touchpad_aspect_ratio":
                            self.aspect_ratio = cast(float, ev["value"])
                            self.touch_correction = correct_touchpad(
                                DS5_EDGE_TOUCH_WIDTH,
                                DS5_EDGE_TOUCH_HEIGHT,
                                self.aspect_ratio,
                                self.touchpad_method,
                            )
                        case "is_attached":
                            new_rep[self.ofs + 52] = (new_rep[self.ofs + 52] & 0x0F) | (
                                0x10 if ev["value"] else 0x00
                            )
                        case "battery":
                            new_rep[self.ofs + 52] = (new_rep[self.ofs + 52] & 0xF0) | (
                                max((ev["value"] - 5) // 10, 0)
                            )

        # Cache
        # if new_rep == self.report:
        #     return
        self.report = new_rep

        #
        # Send report
        #
        # Sequence number
        if new_rep[self.ofs + 6] < 255:
            new_rep[self.ofs + 6] += 1
        else:
            new_rep[self.ofs + 6] = 0

        if self.use_bluetooth:
            sign_crc32_inplace(self.report, DS5_INPUT_CRC32_SEED)
        self.dev.send_input_report(self.report)



import logging
import os
from logging import LogRecord
from logging.handlers import RotatingFileHandler

from .utils import Perms, expanduser, restore_priviledge, switch_priviledge

logger = logging.getLogger(__name__)

RASTER = """\
   _______   _______    ______  
  /    /  ╲╲/    /  ╲╲_/      ╲╲
 /        //        //        //
/         /         /         / 
╲___/____/╲___/____/╲________/  \n"""


class NewLineFormatter(logging.Formatter):
    """Aligns newlines during multiline prints."""

    def format(self, record):
        msg = super().format(record)
        if (idx := msg.index("|||")) != -1:
            preamble = msg[:idx]
            msg = msg.replace("|||", "").replace("\n", "\n" + (" " * len(preamble)))
        return msg


class UserRotatingFileHandler(RotatingFileHandler):
    def __init__(
        self,
        filename: str,
        mode: str = "a",
        maxBytes: int = 0,
        backupCount: int = 0,
        encoding: str | None = None,
        delay: bool = False,
        errors: str | None = None,
        perms: Perms | None = None,
    ) -> None:
        super().__init__(filename, mode, maxBytes, backupCount, encoding, delay, errors)
        self.perms = perms

    def emit(self, record: LogRecord) -> None:
        try:
            # Set permissions to log as user
            if self.perms:
                old = switch_priviledge(self.perms, False)
            else:
                old = None
            RotatingFileHandler.emit(self, record)
            if old:
                restore_priviledge(old)
        except Exception:
            self.handleError(record)


def setup_logger(
    log_dir: str | None = None, init: bool = True, perms: Perms | None = None
):
    from rich import get_console
    from rich.logging import RichHandler
    from rich.traceback import install

    if log_dir:
        log_dir = expanduser(log_dir, perms)

    install()
    handlers = []
    handlers.append(RichHandler())
    if log_dir:
        os.makedirs(log_dir, exist_ok=True)
        handler = UserRotatingFileHandler(
            os.path.join(log_dir, "hhd.log"),
            maxBytes=10_000_000,
            backupCount=10,
            perms=perms,
        )
        handler.setFormatter(
            NewLineFormatter("%(asctime)s %(module)-15s %(levelname)-8s|||%(message)s")
        )
        handler.doRollover()
        handlers.append(handler)

    FORMAT = "%(message)s"
    logging.basicConfig(
        level=logging.INFO,
        datefmt="[%d/%m %H:%M]",
        format=FORMAT,
        handlers=handlers,
    )
    if init:
        get_console().print(RASTER, justify="full", markup=False, highlight=False)
        logger.info(f"Handheld Daemon starting...")



from typing import Any, Sequence

from hhd.plugins import HHDPluginV1, get_relative_fn


def controllers_run(**config: Any):
    from .base import plugin_run

    plugin_run(**config)


def controllers_autodetect():
    # Match just product number, should be enough for now
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        return f.read().strip() == "83E1"


plugins: Sequence[HHDPluginV1] = [
    {
        "name": "legion_go_controllers",
        "autodetect": controllers_autodetect,
        "run": controllers_run,
        "config": get_relative_fn("config.yaml"),
        "config_version": 2,
    }
]


def main():
    from .base import main

    main(False)


if __name__ == "__main__":
    main()



from typing import Any, Sequence

from hhd.plugins import HHDPluginV1, get_relative_fn
from .const import SUPPORTED_DEVICES


def run(**config: Any):
    from .base import power_button_run

    power_button_run(**config)


def autodetect():
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        prod = f.read().strip()

    for d in SUPPORTED_DEVICES:
        if d.prod_name == prod:
            return True

    return False


plugins: Sequence[HHDPluginV1] = [
    {
        "name": "powerbuttond",
        "autodetect": autodetect,
        "run": run,
        "config": None,
    }
]



# Taken from the library hid. TODO: Add proper attribution

import os
import ctypes
import atexit

__all__ = ["HIDException", "DeviceInfo", "Device", "enumerate"]


# hidapi = None
library_paths = (
    "libhidapi-hidraw.so",
    "libhidapi-hidraw.so.0",
    # Only hidraw supported due to the fd requirement
    # "libhidapi-libusb.so",
    # "libhidapi-libusb.so.0",
    # "libhidapi-iohidmanager.so",
    # "libhidapi-iohidmanager.so.0",
    # "libhidapi.dylib",
    # "hidapi.dll",
    # "libhidapi-0.dll",
)

for lib in library_paths:
    try:
        hidapi = ctypes.cdll.LoadLibrary(lib)
        break
    except OSError:
        pass
else:
    error = "Unable to load any of the following libraries:{}".format(
        " ".join(library_paths)
    )
    raise ImportError(error)


hidapi.hid_init()
atexit.register(hidapi.hid_exit)


MAX_REPORT_SIZE = 4096


class HIDException(Exception):
    pass


class DeviceInfo(ctypes.Structure):
    def as_dict(self):
        ret = {}
        for name, type in self._fields_:
            if name == "next":
                continue
            ret[name] = getattr(self, name, None)

        return ret


DeviceInfo._fields_ = [
    ("path", ctypes.c_char_p),
    ("vendor_id", ctypes.c_ushort),
    ("product_id", ctypes.c_ushort),
    ("serial_number", ctypes.c_wchar_p),
    ("release_number", ctypes.c_ushort),
    ("manufacturer_string", ctypes.c_wchar_p),
    ("product_string", ctypes.c_wchar_p),
    ("usage_page", ctypes.c_ushort),
    ("usage", ctypes.c_ushort),
    ("interface_number", ctypes.c_int),
    ("next", ctypes.POINTER(DeviceInfo)),
]


class LinuxHidDevice(ctypes.Structure):
    _fields_ = [
        ("device_handle", ctypes.c_int),
        ("blocking", ctypes.c_int),
        ("last_error_str", ctypes.c_wchar_p),
        ("hid_device_info", ctypes.c_void_p),
    ]


hidapi.hid_init.argtypes = []
hidapi.hid_init.restype = ctypes.c_int
hidapi.hid_exit.argtypes = []
hidapi.hid_exit.restype = ctypes.c_int
hidapi.hid_enumerate.argtypes = [ctypes.c_ushort, ctypes.c_ushort]
hidapi.hid_enumerate.restype = ctypes.POINTER(DeviceInfo)
hidapi.hid_free_enumeration.argtypes = [ctypes.POINTER(DeviceInfo)]
hidapi.hid_free_enumeration.restype = None
hidapi.hid_open.argtypes = [ctypes.c_ushort, ctypes.c_ushort, ctypes.c_wchar_p]
hidapi.hid_open.restype = ctypes.POINTER(LinuxHidDevice)
hidapi.hid_open_path.argtypes = [ctypes.c_char_p]
hidapi.hid_open_path.restype = ctypes.POINTER(LinuxHidDevice)
hidapi.hid_write.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_size_t]
hidapi.hid_write.restype = ctypes.c_int
hidapi.hid_read_timeout.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_size_t,
    ctypes.c_int,
]
hidapi.hid_read_timeout.restype = ctypes.c_int
hidapi.hid_read.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_size_t]
hidapi.hid_read.restype = ctypes.c_int
hidapi.hid_get_input_report.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_size_t,
]
hidapi.hid_get_input_report.restype = ctypes.c_int
hidapi.hid_set_nonblocking.argtypes = [ctypes.c_void_p, ctypes.c_int]
hidapi.hid_set_nonblocking.restype = ctypes.c_int
hidapi.hid_send_feature_report.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
]
hidapi.hid_send_feature_report.restype = ctypes.c_int
hidapi.hid_get_feature_report.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_size_t,
]
hidapi.hid_get_feature_report.restype = ctypes.c_int
hidapi.hid_close.argtypes = [ctypes.c_void_p]
hidapi.hid_close.restype = None
hidapi.hid_get_manufacturer_string.argtypes = [
    ctypes.c_void_p,
    ctypes.c_wchar_p,
    ctypes.c_size_t,
]
hidapi.hid_get_manufacturer_string.restype = ctypes.c_int
hidapi.hid_get_product_string.argtypes = [
    ctypes.c_void_p,
    ctypes.c_wchar_p,
    ctypes.c_size_t,
]
hidapi.hid_get_product_string.restype = ctypes.c_int
hidapi.hid_get_serial_number_string.argtypes = [
    ctypes.c_void_p,
    ctypes.c_wchar_p,
    ctypes.c_size_t,
]
hidapi.hid_get_serial_number_string.restype = ctypes.c_int
hidapi.hid_get_indexed_string.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_wchar_p,
    ctypes.c_size_t,
]
hidapi.hid_get_indexed_string.restype = ctypes.c_int
hidapi.hid_error.argtypes = [ctypes.c_void_p]
hidapi.hid_error.restype = ctypes.c_wchar_p


def enumerate(vid=0, pid=0):
    ret = []
    info = hidapi.hid_enumerate(vid, pid)
    c = info

    while c:
        ret.append(c.contents.as_dict())
        c = c.contents.next

    hidapi.hid_free_enumeration(info)

    return ret


def enumerate_unique(vid=0, pid=0):
    """Returns the current connected devices,
    sorted by path."""
    return sorted(
        list({v["path"]: v for v in enumerate(vid, pid)}.values()),
        key=lambda l: l["path"],
    )


class Device(object):
    def __init__(self, vid=None, pid=None, serial=None, path=None):
        if path:
            self._dev = hidapi.hid_open_path(path)
        elif serial:
            serial = ctypes.create_unicode_buffer(serial)
            self._dev = hidapi.hid_open(vid, pid, serial)
        elif vid and pid:
            self._dev = hidapi.hid_open(vid, pid, None)
        else:
            raise ValueError("specify vid/pid or path")

        if not self._dev:
            raise HIDException("unable to open device")

    @property
    def fd(self):
        if not self._dev:
            return 0
        return self._dev.contents.device_handle

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def __hidcall(self, function, *args, **kwargs):
        if not self._dev:
            raise HIDException("device closed")

        ret = function(*args, **kwargs)

        if ret == -1:
            err = hidapi.hid_error(self._dev)
            raise HIDException(err)
        return ret

    def __readstring(self, function, max_length=255):
        buf = ctypes.create_unicode_buffer(max_length)
        self.__hidcall(function, self._dev, buf, max_length)
        return buf.value

    def write(self, data):
        return self.__hidcall(hidapi.hid_write, self._dev, data, len(data))

    def read(self, size: int = MAX_REPORT_SIZE, timeout=None):
        data = ctypes.create_string_buffer(size)

        if timeout is None:
            size = self.__hidcall(hidapi.hid_read, self._dev, data, size)
        else:
            size = self.__hidcall(
                hidapi.hid_read_timeout, self._dev, data, size, timeout
            )

        return data.raw[:size]

    def get_input_report(self, report_id, size):
        data = ctypes.create_string_buffer(size)

        # Pass the id of the report to be read.
        data[0] = bytearray((report_id,))

        size = self.__hidcall(hidapi.hid_get_input_report, self._dev, data, size)
        return data.raw[:size]

    def send_feature_report(self, data):
        return self.__hidcall(
            hidapi.hid_send_feature_report, self._dev, data, len(data)
        )

    def get_feature_report(self, report_id, size):
        data = ctypes.create_string_buffer(size)

        # Pass the id of the report to be read.
        data[0] = bytearray((report_id,))

        size = self.__hidcall(hidapi.hid_get_feature_report, self._dev, data, size)
        return data.raw[:size]

    def close(self):
        if self._dev:
            hidapi.hid_close(self._dev)
            self._dev = None

    @property
    def nonblocking(self):
        return getattr(self, "_nonblocking", 0)

    @nonblocking.setter
    def nonblocking(self, value):
        self.__hidcall(hidapi.hid_set_nonblocking, self._dev, value)
        setattr(self, "_nonblocking", value)

    @property
    def manufacturer(self):
        return self.__readstring(hidapi.hid_get_manufacturer_string)

    @property
    def product(self):
        return self.__readstring(hidapi.hid_get_product_string)

    @property
    def serial(self):
        return self.__readstring(hidapi.hid_get_serial_number_string)

    def get_indexed_string(self, index, max_length=255):
        buf = ctypes.create_unicode_buffer(max_length)
        self.__hidcall(hidapi.hid_get_indexed_string, self._dev, index, buf, max_length)
        return buf.value



import logging
import os
import subprocess
from typing import NamedTuple

logger = logging.getLogger(__name__)


class Perms(NamedTuple):
    euid: int = 0
    egid: int = 0
    uid: int = 0
    gid: int = 0
    name: str = "root"


def get_perms(user: str | None) -> Perms | None:
    try:
        uid = os.getuid()
        gid = os.getgid()

        if not user:
            if not uid or not gid:
                print(f"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                print(
                    "Running as root without a specified user (`--user`). Configs will be placed at `/root/.config`."
                )
                print(f"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            return Perms(uid, gid, uid, gid, "root")

        euid = int(
            subprocess.run(
                ["id", "-u", user], capture_output=True, check=True
            ).stdout.decode()
        )
        egid = int(
            subprocess.run(
                ["id", "-g", user], capture_output=True, check=True
            ).stdout.decode()
        )

        if (uid or gid) and (uid != euid or gid != egid):
            print(
                f"The user specified with --user is not the user this process was started with."
            )
            return None

        return Perms(euid, egid, uid, gid, user)
    except subprocess.CalledProcessError as e:
        print(f"Getting the user uid/gid returned an error:\n{e.stderr.decode()}")
        return None
    except Exception as e:
        print(f"Failed getting permissions with error:\n{e}")
        return None


def switch_priviledge(p: Perms, escalate=False):
    uid = os.geteuid()
    gid = os.getegid()

    if escalate:
        os.seteuid(p.uid)
        os.setegid(p.gid)
    else:
        os.setegid(p.egid)
        os.seteuid(p.euid)

    return uid, gid


def restore_priviledge(old: tuple[int, int]):
    uid, gid = old
    # Try writing group first in case of root
    # and fail silently
    try:
        os.setegid(gid)
    except Exception:
        pass
    os.seteuid(uid)
    os.setegid(gid)
    pass


def expanduser(path: str, user: int | str | Perms | None = None):
    """Expand ~ and ~user constructions.  If user or $HOME is unknown,
    do nothing.

    Modified from the python implementation to support using the target userid/user."""

    path = os.fspath(path)

    if not path.startswith("~"):
        return path

    i = path.find("/", 1)
    if i < 0:
        i = len(path)
    if i == 1:
        if "HOME" in os.environ and not user:
            # Fallback to environ only if user not set
            userhome = os.environ["HOME"]
        else:
            try:
                import pwd
            except ImportError:
                # pwd module unavailable, return path unchanged
                return path
            try:
                if not user:
                    userhome = pwd.getpwuid(os.getuid()).pw_dir
                elif isinstance(user, int):
                    userhome = pwd.getpwuid(user).pw_dir
                elif isinstance(user, Perms):
                    userhome = pwd.getpwuid(user.euid).pw_dir
                else:
                    userhome = pwd.getpwnam(user).pw_dir
            except KeyError:
                # bpo-10496: if the current user identifier doesn't exist in the
                # password database, return the path unchanged
                return path
    else:
        try:
            import pwd
        except ImportError:
            # pwd module unavailable, return path unchanged
            return path
        name = path[1:i]
        try:
            pwent = pwd.getpwnam(name)
        except KeyError:
            # bpo-10496: if the user name from the path doesn't exist in the
            # password database, return the path unchanged
            return path
        userhome = pwent.pw_dir

    root = "/"
    userhome = userhome.rstrip(root)
    return (userhome + path[i:]) or root



import argparse
import logging
import os
import select
import shutil
import subprocess
import time
from multiprocessing import Process
from os.path import join
from typing import NamedTuple

import pkg_resources
import yaml

from .logging import setup_logger
from .plugins import HHDPluginV1
from .utils import Perms, expanduser, get_perms, switch_priviledge

logger = logging.getLogger(__name__)

CONFIG_DIR = os.environ.get("HHD_CONFIG_DIR", "~/.config/hhd")

ERROR_DELAY = 5


def launch_plugin(pkg_name: str, plugin: HHDPluginV1, perms: Perms):
    plugin_dir = expanduser(join(CONFIG_DIR, "plugins"), perms)

    if plugin["config"]:
        cfg_fn = join(plugin_dir, plugin["name"] + ".yaml")

        if not plugin["autodetect"]():
            logger.debug(f"Plugin '{plugin['name']}' determined it should not run.")
            return None

        if not os.path.isfile(cfg_fn):
            logger.warn(
                f"Config file for plugin '{plugin['name']}' not found below, using default:\n{cfg_fn}"
            )
            os.makedirs(plugin_dir, exist_ok=True)
            shutil.copy(plugin["config"], cfg_fn)

        with open(cfg_fn, "r") as f:
            cfg = yaml.safe_load(f)

        if "config_version" in plugin and plugin["config_version"] != cfg.get(
            "version", 0
        ):
            logger.warn(
                f"Config file for plugin '{plugin['name']}' is outdated, replacing with default:\n{cfg_fn}"
            )
            os.makedirs(plugin_dir, exist_ok=True)
            shutil.copy(plugin["config"], cfg_fn)

            with open(cfg_fn, "r") as f:
                cfg = yaml.safe_load(f)
    else:
        cfg = {}

    # Add perms in case a plugin needs them
    cfg = {**cfg, "perms": perms}

    logger.info(f"Launching plugin {pkg_name}.{plugin['name']}")

    # Run plugin priviledged, in case it does not deal with the user
    switch_priviledge(perms, True)
    proc = Process(target=plugin["run"], kwargs=cfg)
    proc.start()
    switch_priviledge(perms, False)

    return proc


def main():
    # Set up permissions
    parser = argparse.ArgumentParser(
        prog="HHD: Handheld Daemon main interface.",
        description="Handheld Daemon is a daemon for managing the quirks inherent in handheld devices.",
    )
    parser.add_argument(
        "-u",
        "--user",
        default=None,
        help="The user whose home directory will be used to store the files (~/.config/hhd).",
        dest="user",
    )
    args = parser.parse_args()
    user = args.user

    # Setup temporary logger for permission retreival
    perms = get_perms(user)
    if not perms:
        print(f"Could not get user information. Exiting...")
        return

    running_plugins: dict[int, tuple[str, HHDPluginV1, Process]] = {}
    try:
        # Drop privileges for initial set-up
        switch_priviledge(perms, False)
        setup_logger(
            join(CONFIG_DIR, "log"), perms=perms
        )

        plugins = {}
        for plugin in pkg_resources.iter_entry_points("hhd.plugins"):
            plugins[plugin.name] = plugin.resolve()

        logger.info(f"Found plugin providers: {', '.join(list(plugins))}")
        plugin_str = "With the following plugins:"
        for pkg_name, sub_plugins in plugins.items():
            plugin_str += (
                f"\n  - {pkg_name:>15s}: {', '.join(p['name'] for p in sub_plugins)}"
            )
        logger.info(plugin_str)

        for pkg_name, sub_plugins in plugins.items():
            for plugin in sub_plugins:
                proc = launch_plugin(pkg_name, plugin, perms)
                if proc:
                    running_plugins[proc.sentinel] = (pkg_name, plugin, proc)

        if not running_plugins:
            logger.error(f"No plugins started, exiting...")
            return

        logger.info(f"Monitoring plugin status, and restarting if necessary.")
        while True:
            exited = select.select(list(running_plugins), [], [])[0]
            for fd in exited:
                pkg_name, plugin, proc = running_plugins.pop(fd)
                if not proc.exitcode:
                    # Plugin exited normally, not restarting
                    logger.info(f"Plugin '{plugin['name']}' exited normally.")
                    continue

                logger.error(
                    f"Plugin '{plugin['name']}' crashed. Restarting in {ERROR_DELAY}s."
                )
                time.sleep(ERROR_DELAY)
                proc = launch_plugin(pkg_name, plugin, perms)
                if proc:
                    running_plugins[proc.sentinel] = (pkg_name, plugin, proc)
            time.sleep(ERROR_DELAY)
    except KeyboardInterrupt:
        logger.info(
            f"HHD Daemon received KeyboardInterrupt, stopping plugins and exiting."
        )
    finally:
        for _, _, process in running_plugins.values():
            process.terminate()
            process.join()


if __name__ == "__main__":
    main()



from typing import Any, Protocol, TypedDict


def get_relative_fn(fn: str):
    """Returns the directory of a file relative to the script calling this function."""
    import inspect
    import os

    script_fn = inspect.currentframe().f_back.f_globals["__file__"]  # type: ignore
    dirname = os.path.dirname(script_fn)
    return os.path.join(dirname, fn)


class HHDPluginV1Autodetect(Protocol):
    def __call__(self) -> bool:
        return True


class HHDPluginV1Run(Protocol):
    def __call__(self, **config: Any) -> None:
        pass


class HHDPluginV1Nonversioned(TypedDict):
    name: str
    autodetect: HHDPluginV1Autodetect
    run: HHDPluginV1Run
    config: str | None


class HHDPluginV1Versioned(TypedDict):
    name: str
    autodetect: HHDPluginV1Autodetect
    run: HHDPluginV1Run
    config: str | None
    config_version: int


HHDPluginV1 = HHDPluginV1Nonversioned | HHDPluginV1Versioned
"""Initial version of HHD plugins. These plugins use static configuration,
and it is not possible for them to share configuration with each other
of communicate through d-bus. Restarting hhd is required for reloading
configuration.

The plugins specify a name, a run function, and the location of a template
config file. Use the function `get_relative_fn()` to retrieve the directory
of the script executing it, and place the config file relative to it.

To stop the plugin, a KeyboardInterrupt is raised.

Provide `config_version` to replace your config file on startup when a version
changes.

@warning: `run()` needs to be a global function, e.g., it can not
be a lambda and decorators are tricky."""



from threading import Thread, Event
import time
from hhd.controller.physical.imu import ForcedSampler

import logging

logger = logging.getLogger(__name__)


def gyro_fix(ev: Event, rate: int = 65):
    g = None
    try:
        g = ForcedSampler(["gyro_3d"], False)
        g.open()
        while not ev.is_set():
            g.sample()
            time.sleep(1 / rate)
    except KeyboardInterrupt:
        raise
    except Exception as e:
        logger.warn(f"Gyro fix failed with error:{e}\n")
    finally:
        if g:
            g.close()

class GyroFixer:
    def __init__(self, rate: int = 65) -> None:
        self.rate = rate
        self.ev = None
        self.thread = None

    def open(self):
        logger.info("Starting gyro fixer thread.")
        self.close()
        self.ev = Event()
        self.thread = Thread(target=gyro_fix, args=(self.ev, self.rate))
        self.thread.start()

    def close(self):
        if self.ev:
            logger.info("Stopping the gyro fixer thread.")
            self.ev.set()
        if self.thread:
            self.thread.join()
        self.ev = None
        self.thread = None



import select
from typing import Any, Literal, Sequence, TypedDict
import time
from .const import Axis, Button, Configuration


class EnvelopeEffect(TypedDict):
    type: Literal["constant", "ramp"]
    attack_length: int
    attack_level: int
    fade_length: int
    fade_level: int


class ConditionSide(TypedDict):
    # TODO: replace with enums
    right_saturation: int
    left_saturation: int
    right_coeff: int
    left_coeff: int
    deadband: int
    center: int


class RumbleEffect(TypedDict):
    type: Literal["rumble"]
    strong_magnitude: float
    weak_magnitude: float


class ConditionEffect(TypedDict):
    type: Literal["condition"]
    left: ConditionSide
    right: ConditionSide


class PeriodicEffect(TypedDict):
    type: Literal["periodic"]
    # Todo: replace with enums
    waveform: int
    period: int
    magnitude: int
    offset: int
    phase: int

    attack_length: int
    attack_level: int
    fade_length: int
    fade_level: int

    custom: bytes


class EffectEvent(TypedDict):
    # Always effect, better for filterring
    type: Literal["effect"]
    # Event target. Not part of the standard but required for e.g., DS5.
    code: Literal["main", "left", "right"]

    id: int
    # TODO: Upgrade to literal
    direction: int

    trigger_button: str
    trigger_interval: int

    replay_length: int
    replay_delay: int

    effect: EnvelopeEffect | ConditionEffect | PeriodicEffect | RumbleEffect


class RumbleEvent(TypedDict):
    """In case ev effects is too complicated. If both magnitudes are 0, disable rumble."""

    type: Literal["rumble"]
    code: Literal["main", "left", "right"]
    strong_magnitude: float
    weak_magnitude: float


class RgbLedEvent(TypedDict):
    """Inspired by new controllers with RGB leds, especially below the buttons.

    Instead of code, this event type exposes multiple properties, including mode."""

    type: Literal["led"]

    # The led
    code: Literal["main", "left", "right"]

    # Various lighting modes supported by the led.
    mode: Literal["disable", "solid", "blinking", "rainbow", "spiral"]

    # Brightness range is from 0 to 1
    # If the response device does not support brightness control, it shall
    # convert the rgb color to hue, assume saturation is 1, and derive a new
    # RGB value from the brightness below
    brightness: float

    # The speed the led should blink if supported by the led
    speed: float

    # Color values for the led, may be ommited depending on the mode, by being
    # set to 0
    red: int
    green: int
    blue: int


class ButtonEvent(TypedDict):
    type: Literal["button"]
    code: Button
    value: bool


class AxisEvent(TypedDict):
    type: Literal["axis"]
    code: Axis
    value: float


class ConfigurationEvent(TypedDict):
    type: Literal["configuration"]
    code: Configuration
    value: Any


Event = (
    EffectEvent
    | ButtonEvent
    | AxisEvent
    | ConfigurationEvent
    | RgbLedEvent
    | RumbleEvent
)


class Producer:
    def open(self) -> Sequence[int]:
        """Opens and returns a list of file descriptors that should be listened to."""
        raise NotImplementedError()

    def close(self, exit: bool) -> bool:
        """Called to close the device.

        If `exit` is true, the program is about to
        close. If it is false, the controller is entering power save mode because
        it is unused. In this case, if this service is required, you may forgo
        closing and return false. If true, it is assumed this producer is closed.

        `open()` will be called again once the consumers are ready."""
        return False

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        """Called with the file descriptors that are ready to read."""
        return []


class Consumer:
    available: bool
    """Hint that states if the consumer can receive events. If it is false,
    consumer will not be called. If all consumers are false, producers will
    be closed to save CPU utilisation."""

    def initialize(self):
        """Optional method for initialization."""
        pass

    def consume(self, events: Sequence[Event]):
        pass


class Multiplexer:
    QAM_DELAY = 0.2

    def __init__(
        self,
        swap_guide: None | Literal["guide_is_start", "guide_is_select"] = None,
        trigger: None | Literal["analog_to_discrete", "discrete_to_analogue"] = None,
        dpad: None | Literal["analog_to_discrete"] = None,
        led: None | Literal["left_to_main", "right_to_main", "main_to_sides"] = None,
        touchpad: None
        | Literal["left_to_main", "right_to_main", "main_to_sides"] = None,
        status: None | Literal["both_to_main"] = None,
        share_to_qam: bool = False,
        trigger_discrete_lvl: float = 0.99,
    ) -> None:
        self.swap_guide = swap_guide
        self.trigger = trigger
        self.dpad = dpad
        self.led = led
        self.touchpad = touchpad
        self.status = status
        self.trigger_discrete_lvl = trigger_discrete_lvl
        self.share_to_qam = share_to_qam

        self.state = {}
        self.queue: list[tuple[Event, int]] = []

        assert touchpad is None, "touchpad rewiring not supported yet"

    def process(self, events: Sequence[Event]):
        out: list[Event] = []
        status_events = set()

        curr = time.perf_counter()
        while len(self.queue) and self.queue[0][1] < curr:
            out.append(self.queue.pop(0)[0])

        for ev in events:
            match ev["type"]:
                case "axis":
                    if self.trigger == "analog_to_discrete" and ev["code"] in (
                        "lt",
                        "rt",
                    ):
                        out.append(
                            {
                                "type": "button",
                                "code": ev["code"],
                                "value": ev["value"] > self.trigger_discrete_lvl,
                            }
                        )

                    if self.dpad == "analog_to_discrete" and ev["code"] in (
                        "hat_x",
                        "hat_y",
                    ):
                        out.append(
                            {
                                "type": "button",
                                "code": "dpad_up"
                                if ev["code"] == "hat_y"
                                else "dpad_right",
                                "value": ev["value"] > 0.5,
                            }
                        )
                        out.append(
                            {
                                "type": "button",
                                "code": "dpad_down"
                                if ev["code"] == "hat_y"
                                else "dpad_left",
                                "value": ev["value"] < -0.5,
                            }
                        )
                case "button":
                    if self.trigger == "discrete_to_analog" and ev["code"] in (
                        "lt",
                        "rt",
                    ):
                        out.append(
                            {
                                "type": "axis",
                                "code": ev["code"],
                                "value": 1 if ev["value"] else 0,
                            }
                        )

                    if self.swap_guide and ev["code"] in (
                        "start",
                        "select",
                        "mode",
                        "share",
                    ):
                        match ev["code"]:
                            case "start":
                                ev["code"] = "mode"
                            case "select":
                                ev["code"] = "share"
                            case "mode":
                                if self.swap_guide == "guide_is_start":
                                    ev["code"] = "start"
                                else:
                                    ev["code"] = "select"
                            case "share":
                                if self.swap_guide == "guide_is_start":
                                    ev["code"] = "select"
                                else:
                                    ev["code"] = "start"

                    if self.share_to_qam and ev["code"] == "share":
                        if ev["value"]:
                            ev["code"] = "mode"
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "a",
                                        "value": True,
                                    },
                                    curr + self.QAM_DELAY,
                                )
                            )
                        else:
                            # TODO: Clean this up
                            ev["code"] = ""  # type: ignore
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "mode",
                                        "value": False,
                                    },
                                    curr + self.QAM_DELAY,
                                ),
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "a",
                                        "value": False,
                                    },
                                    curr + self.QAM_DELAY,
                                ),
                            )

                        # append A after QAM_DELAY s
                case "led":
                    if self.led == "left_to_main" and ev["code"] == "left":
                        out.append({**ev, "code": "main"})
                    elif self.led == "right_to_main" and ev["code"] == "right":
                        out.append({**ev, "code": "main"})
                    elif self.led == "main_to_both" and ev["code"] == "main":
                        out.append({**ev, "code": "left"})
                        out.append({**ev, "code": "right"})
                case "configuration":
                    if self.status == "both_to_main":
                        self.state[ev["code"]] = ev["value"]
                        match ev["code"]:
                            case "battery_left" | "battery_right":
                                status_events.add("battery")
                            case "is_attached_left" | "is_attached_right":
                                status_events.add("is_attached")
                            case "is_connected_left" | "is_connected_right":
                                status_events.add("is_connected")

        for s in status_events:
            match s:
                case "battery":
                    out.append(
                        {
                            "type": "configuration",
                            "code": "battery",
                            "value": min(
                                self.state.get("battery_left", 0),
                                self.state.get("battery_right", 0),
                            ),
                        }
                    )
                case "is_attached":
                    out.append(
                        {
                            "type": "configuration",
                            "code": "is_attached",
                            "value": (
                                self.state.get("is_attached_left", False)
                                and self.state.get("is_attached_right", False)
                            ),
                        }
                    )
                case "is_connected":
                    out.append(
                        {
                            "type": "configuration",
                            "code": "is_connected",
                            "value": (
                                self.state.get("is_connected_left", False)
                                and self.state.get("is_connected_right", False)
                            ),
                        }
                    )

        out.extend(events)
        return out


def can_read(fd: int):
    return select.select([fd], [], [], 0)[0]



import logging
from typing import Sequence, cast

import evdev
from evdev import UInput

from hhd.controller import Axis, Button, Consumer, Producer
from hhd.controller.base import Event, can_read


def B(b: str):
    return cast(int, getattr(evdev.ecodes, b))


logger = logging.getLogger(__name__)

GAMEPAD_BTN_CAPABILITIES = {
    B("EV_KEY"): [
        B("BTN_TL"),
        B("BTN_TR"),
        B("BTN_SELECT"),
        B("BTN_START"),
        B("BTN_MODE"),
        B("BTN_THUMBL"),
        B("BTN_THUMBR"),
        B("BTN_A"),
        B("BTN_B"),
        B("BTN_X"),
        B("BTN_Y"),
        B("BTN_MODE"),
        B("BTN_TRIGGER_HAPPY1"),
        B("BTN_TRIGGER_HAPPY2"),
        B("BTN_TRIGGER_HAPPY3"),
        B("BTN_TRIGGER_HAPPY4"),
        B("BTN_TRIGGER_HAPPY5"),
        B("BTN_TRIGGER_HAPPY6"),
    ]
}
STANDARD_BUTTON_MAP: dict[Button, int] = {
    # Gamepad
    "a": B("BTN_A"),
    "b": B("BTN_B"),
    "x": B("BTN_X"),
    "y": B("BTN_Y"),
    # Sticks
    "ls": B("BTN_THUMBL"),
    "rs": B("BTN_THUMBR"),
    # Bumpers
    "lb": B("BTN_TL"),
    "rb": B("BTN_TR"),
    # Select
    "start": B("BTN_START"),
    "select": B("BTN_SELECT"),
    # Misc
    "mode": B("BTN_MODE"),
    # Back buttons
    "extra_l1": B("BTN_TRIGGER_HAPPY1"),
    "extra_l2": B("BTN_TRIGGER_HAPPY2"),
    "extra_l3": B("BTN_TRIGGER_HAPPY5"),
    "extra_r1": B("BTN_TRIGGER_HAPPY3"),
    "extra_r2": B("BTN_TRIGGER_HAPPY4"),
    "extra_r3": B("BTN_TRIGGER_HAPPY6"),
}


class UInputDevice(Consumer, Producer):
    def __init__(
        self,
        capabilities=GAMEPAD_BTN_CAPABILITIES,
        btn_map: dict[Button, int] = STANDARD_BUTTON_MAP,
        axis_map: dict[Axis, int] = {},
        vid: int = 2,
        pid: int = 2,
        name: str = "HHD Shortcuts Device",
    ) -> None:
        self.capabilities = capabilities
        self.btn_map = btn_map
        self.axis_map = axis_map
        self.dev = None
        self.name = name
        self.vid = vid
        self.pid = pid

    def open(self) -> Sequence[int]:
        logger.info(f"Opening virtual device '{self.name}'")
        self.dev = UInput(
            events=self.capabilities, name=self.name, vendor=self.vid, product=self.pid
        )
        self.fd = self.dev.fd
        return [self.fd]

    def close(self, exit: bool) -> bool:
        if self.dev:
            self.dev.close()
        self.input = None
        self.fd = None
        return True

    def consume(self, events: Sequence[Event]):
        if not self.dev:
            return
        for ev in events:
            match ev["type"]:
                case "axis":
                    # if ev["code"] in self.axis_map:
                    #     self.dev.write(B("EV_ABS"), self.axis_map[ev["code"]], ev['value'])
                    # TODO: figure out normalization
                    if ev['value']:
                        logger.error(f"Outputing axis not supported yet. Event:\n{ev}")
                case "button":
                    if ev["code"] in self.btn_map:
                        self.dev.write(
                            B("EV_KEY"),
                            self.btn_map[ev["code"]],
                            1 if ev["value"] else 0,
                        )
        self.dev.syn()



# SPDX-License-Identifier: MIT
# Taken from unmaintained library python-uhid, mostly rewritten TODO: Add proper attribution

from __future__ import annotations

import enum
import os
import os.path
import select
import struct
import sys
import uuid
from typing import Literal, Optional, TypedDict

from hhd.controller import can_read

# _HID_MAX_DESCRIPTOR_SIZE = 4096
UHID_DATA_MAX = 4096


BUS_PCI = 0x01
BUS_ISAPNP = 0x02
BUS_USB = 0x03
BUS_HIL = 0x04
BUS_BLUETOOTH = 0x05
BUS_VIRTUAL = 0x06


# UHID_LEGACY_CREATE = 0
UHID_DESTROY = 1
UHID_START = 2
UHID_STOP = 3
UHID_OPEN = 4
UHID_CLOSE = 5
UHID_OUTPUT = 6
# UHID_LEGACY_OUTPUT_EV = 7
# UHID_LEGACY_INPUT = 8
UHID_GET_REPORT = 9
UHID_GET_REPORT_REPLY = 10
UHID_CREATE2 = 11
UHID_INPUT2 = 12
UHID_SET_REPORT = 13
UHID_SET_REPORT_REPLY = 14


class DevFlag(enum.Enum):
    UHID_DEV_NUMBERED_FEATURE_REPORTS = 1 << 0
    UHID_DEV_NUMBERED_OUTPUT_REPORTS = 1 << 1
    UHID_DEV_NUMBERED_INPUT_REPORTS = 1 << 2


class ReportType(enum.Enum):
    UHID_FEATURE_REPORT = 0
    UHID_OUTPUT_REPORT = 1
    UHID_INPUT_REPORT = 2


# Used as a reference
# class Create2Req(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = [
#         ("name", ctypes.c_char * 128),
#         ("phys", ctypes.c_char * 64),
#         ("uniq", ctypes.c_char * 64),
#         ("rd_size", ctypes.c_uint16),
#         ("bus", ctypes.c_uint16),
#         ("vendor", ctypes.c_uint32),
#         ("product", ctypes.c_uint32),
#         ("version", ctypes.c_uint32),
#         ("country", ctypes.c_uint32),
#         ("rd_data", ctypes.c_char * _HID_MAX_DESCRIPTOR_SIZE),
#     ]


# class StartReq(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = [
#         ("dev_flags", ctypes.c_uint64),
#     ]


# class Input2Req(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = [
#         ("size", ctypes.c_uint16),
#         ("data", ctypes.c_char * _UHID_DATA_MAX),
#     ]


# class OutputReq(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = [
#         ("data", ctypes.c_uint8 * _UHID_DATA_MAX),
#         ("size", ctypes.c_uint16),
#         ("rtype", ctypes.c_uint8),
#     ]


# class GetReportReq(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = [
#         ("id", ctypes.c_uint32),
#         ("rnum", ctypes.c_uint8),
#         ("rtype", ctypes.c_uint8),
#     ]


# class GetReportReplyReq(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = [
#         ("id", ctypes.c_uint32),
#         ("err", ctypes.c_uint16),
#         ("size", ctypes.c_uint16),
#         ("data", ctypes.c_char * _UHID_DATA_MAX),
#     ]


# class SetReportReq(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = [
#         ("id", ctypes.c_uint64),
#         ("rnum", ctypes.c_uint8),
#         ("rtype", ctypes.c_uint8),
#         ("size", ctypes.c_uint16),
#         ("data", ctypes.c_char * _UHID_DATA_MAX),
#     ]


# class SetReportReplyReq(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = [
#         ("id", ctypes.c_uint32),
#         ("err", ctypes.c_uint16),
#     ]


# class DestroyReq(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = []


# class _U(ctypes.Union):
#     _fields_ = [
#         ("output", OutputReq),
#         ("get_report", GetReportReq),
#         ("get_report_reply", GetReportReplyReq),
#         ("create2", Create2Req),
#         ("input2", Input2Req),
#         ("set_report", SetReportReq),
#         ("set_report_reply", SetReportReplyReq),
#         ("start", StartReq),
#     ]


# class Event(ctypes.Structure):
#     _pack_ = 1
#     _fields_ = [
#         ("type", ctypes.c_uint32),
#         ("u", _U),
#     ]


class EventStart(TypedDict):
    type: Literal["start"]
    dev_flags: int


class EventOutput(TypedDict):
    type: Literal["output"]
    report: int
    data: bytes


class EventGetReport(TypedDict):
    type: Literal["get_report"]
    id: int
    rnum: int
    rtype: int


class EventSetReport(TypedDict):
    type: Literal["set_report"]
    id: int
    rnum: int
    rtype: int
    data: bytes


class EventOther(TypedDict):
    type: Literal["open", "close", "stop"]


class UhidDevice:
    def __init__(
        self,
        vid: int,
        pid: int,
        name: bytes,
        report_descriptor: bytes,
        bus: int = BUS_USB,
        physical_name: Optional[bytes] = None,
        unique_name: Optional[bytes] = None,
        version: int = 0,
        country: int = 0,
    ) -> None:
        if not unique_name:
            unique_name = f"{self.__class__.__name__}_{uuid.uuid4()}"[:63].encode()

        if not physical_name:
            physical_name = f"{self.__class__.__name__}/{unique_name}"[:63].encode()

        self.bus = bus
        self.vid = vid
        self.pid = pid
        self.name = name
        self.physical_name = physical_name
        self.unique_name = unique_name
        self.version = version
        self.country = country
        self.report_descriptor = report_descriptor

        self.fd = 0
        self.poll = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(vid={self.vid}, pid={self.pid}, name={self.name}, uniq={self.unique_name})"

    def open(self):
        self.send_create()
        return self.fd

    def close(self):
        if self.fd:
            os.close(self.fd)
            self.fd = 0

    def send_event(self, event: bytes):
        if not self.fd:
            self.fd = os.open("/dev/uhid", os.O_RDWR)
        os.write(self.fd, event)

    def read_event(
        self,
    ) -> None | EventOther | EventStart | EventOutput | EventSetReport | EventGetReport:
        if not self.fd or not can_read(self.fd):
            return None

        # + 4 for desc, + 3 for output report
        d = os.read(self.fd, UHID_DATA_MAX + 4 + 3)

        v = int.from_bytes(d[:4], byteorder=sys.byteorder)
        if v == UHID_START:
            return {
                "type": "start",
                "dev_flags": int.from_bytes(d[4:12], byteorder=sys.byteorder),
            }
        elif v == UHID_STOP:
            return {"type": "stop"}
        elif v == UHID_OPEN:
            return {"type": "open"}
        elif v == UHID_CLOSE:
            return {"type": "close"}
        elif v == UHID_OUTPUT:
            l = int.from_bytes(d[-3:-1], byteorder=sys.byteorder)
            return {"type": "output", "report": d[-1], "data": d[4 : 4 + l]}
        elif v == UHID_SET_REPORT:
            return {
                "type": "set_report",
                "id": int.from_bytes(d[4:8], byteorder=sys.byteorder),
                "rnum": d[8],
                "rtype": d[9],
                "data": d[10:],
            }
        elif v == UHID_GET_REPORT:
            return {
                "type": "get_report",
                "id": int.from_bytes(d[4:8], byteorder=sys.byteorder),
                "rnum": d[8],
                "rtype": d[9],
            }
        assert False, f"Report type {v} uknown"

    def send_create(self) -> None:
        ev = (
            struct.pack(
                "< L 128s 64s 64s H H L L L L",
                UHID_CREATE2,
                self.name,
                self.physical_name,
                self.unique_name,
                len(self.report_descriptor),
                self.bus,
                self.vid,
                self.pid,
                self.version,
                self.country,
            )
            + self.report_descriptor
        )
        self.send_event(ev)

    def send_destroy(self) -> None:
        self.send_event(int.to_bytes(UHID_DESTROY, 4, byteorder=sys.byteorder))

    def send_input_report(self, data: bytes):
        ev = struct.pack("< L H", UHID_INPUT2, len(data)) + data
        self.send_event(ev)

    def send_get_report_reply(self, id: int, err: int, data: bytes):
        ev = struct.pack("< L L H H", UHID_GET_REPORT_REPLY, id, err, len(data)) + data
        self.send_event(ev)

    def send_set_report_reply(self, id: int, err: int):
        ev = struct.pack("< L L H", UHID_GET_REPORT_REPLY, id, err)
        self.send_event(ev)



import logging
import re
import select
from typing import Mapping, Sequence, TypeVar, cast

import evdev
from evdev import ff, ecodes

from hhd.controller import Axis, Button, Event, Producer, Consumer
from hhd.controller.base import Event
from hhd.controller.lib.common import hexify, matches_patterns

from ..const import AbsAxis, GamepadButton, KeyboardButton

logger = logging.getLogger(__name__)


def B(b: str):
    return cast(int, getattr(evdev.ecodes, b))


A = TypeVar("A")


def to_map(b: dict[A, Sequence[int]]) -> dict[int, A]:
    out = {}
    for btn, seq in b.items():
        for s in seq:
            out[s] = btn
    return out


XBOX_BUTTON_MAP: dict[int, GamepadButton] = to_map(
    {
        # Gamepad
        "a": [B("BTN_A")],
        "b": [B("BTN_B")],
        "x": [B("BTN_X")],
        "y": [B("BTN_Y")],
        # Sticks
        "ls": [B("BTN_THUMBL")],
        "rs": [B("BTN_THUMBR")],
        # Bumpers
        "lb": [B("BTN_TL")],
        "rb": [B("BTN_TR")],
        # Select
        "start": [B("BTN_START")],
        "select": [B("BTN_SELECT")],
        # Misc
        "mode": [B("BTN_MODE")],
    }
)

XBOX_AXIS_MAP: dict[int, AbsAxis] = to_map(
    {
        # Sticks
        # Values should range from -1 to 1
        "ls_x": [B("ABS_X")],
        "ls_y": [B("ABS_Y")],
        "rs_x": [B("ABS_RX")],
        "rs_y": [B("ABS_RY")],
        # Triggers
        # Values should range from -1 to 1
        "rt": [B("ABS_Z")],
        "lt": [B("ABS_RZ")],
        # Hat, implemented as axis. Either -1, 0, or 1
        "hat_x": [B("ABS_HAT0X")],
        "hat_y": [B("ABS_HAT0Y")],
    }
)


class GenericGamepadEvdev(Producer, Consumer):
    def __init__(
        self,
        vid: Sequence[int],
        pid: Sequence[int],
        name: Sequence[str | re.Pattern],
        btn_map: Mapping[int, Button] = XBOX_BUTTON_MAP,
        axis_map: Mapping[int, Axis] = XBOX_AXIS_MAP,
        aspect_ratio: float | None = None,
        required: bool = True,
    ) -> None:
        self.vid = vid
        self.pid = pid
        self.name = name

        self.btn_map = btn_map
        self.axis_map = axis_map
        self.aspect_ratio = aspect_ratio

        self.dev: evdev.InputDevice | None = None
        self.fd = 0
        self.required = required

    def open(self) -> Sequence[int]:
        for d in evdev.list_devices():
            dev = evdev.InputDevice(d)
            if not matches_patterns(dev.info.vendor, self.vid):
                continue
            if not matches_patterns(dev.info.product, self.pid):
                continue
            if not matches_patterns(dev.name, self.name):
                continue
            self.dev = dev
            self.dev.grab()
            self.ranges = {
                a: (i.min, i.max) for a, i in self.dev.capabilities().get(B("EV_ABS"), [])  # type: ignore
            }
            self.fd = dev.fd
            self.started = True
            self.effect_id = -1
            return [self.fd]

        err = f"Device with the following not found:\n"
        if self.vid:
            err += f"Vendor ID: {hexify(self.vid)}\n"
        if self.pid:
            err += f"Product ID: {hexify(self.pid)}\n"
        if self.name:
            err += f"Name: {self.name}\n"
        logger.error(err)
        if self.required:
            raise RuntimeError()
        return []

    def close(self, exit: bool) -> bool:
        if self.dev:
            self.dev.close()
            self.fd = 0
        return True

    def consume(self, events: Sequence[Event]):
        if not self.dev:
            return

        for ev in events:
            match ev["type"]:
                case "rumble":
                    # Erase old effect
                    if self.effect_id != -1:
                        self.dev.erase_effect(self.effect_id)
                        self.effect_id = -1

                    # Install new effect
                    if ev["strong_magnitude"] > 0 or ev["weak_magnitude"] > 0:
                        rumble = ff.Rumble(
                            strong_magnitude=0x0000, weak_magnitude=0xFFFF
                        )
                        duration_ms = 10000

                        effect = ff.Effect(
                            getattr(ecodes, "FF_RUMBLE"),
                            -1,
                            0,
                            ff.Trigger(0, 0),
                            ff.Replay(duration_ms, 0),
                            ff.EffectType(ff_rumble_effect=rumble),
                        )
                        self.effect_id = self.dev.upload_effect(effect)
                        self.dev.write(getattr(ecodes, "EV_FF"), self.effect_id, 1)

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if not self.dev or not self.fd in fds:
            return []

        out: list[Event] = []
        if self.started and self.aspect_ratio is not None:
            self.started = False
            out.append(
                {
                    "type": "configuration",
                    "code": "touchpad_aspect_ratio",
                    "value": self.aspect_ratio,
                }
            )

        while select.select([self.fd], [], [], 0)[0]:
            for e in self.dev.read():
                if e.type == B("EV_KEY"):
                    if e.code in self.btn_map:
                        out.append(
                            {
                                "type": "button",
                                "code": self.btn_map[e.code],
                                "value": bool(e.value),
                            }
                        )
                elif e.type == B("EV_ABS"):
                    if e.code in self.axis_map:
                        # Normalize
                        val = e.value / abs(
                            self.ranges[e.code][1 if e.value >= 0 else 0]
                        )

                        out.append(
                            {
                                "type": "axis",
                                "code": self.axis_map[e.code],
                                "value": val,
                            }
                        )
        return out


KEYBOARD_MAP: dict[int, KeyboardButton] = to_map(
    {
        "key_esc": [B("KEY_ESC")],  # 1
        "key_enter": [B("KEY_ENTER")],  # 28
        "key_leftctrl": [B("KEY_LEFTCTRL")],  # 29
        "key_leftshift": [B("KEY_LEFTSHIFT")],  # 42
        "key_leftalt": [B("KEY_LEFTALT")],  # 56
        "key_rightctrl": [B("KEY_RIGHTCTRL")],  # 97
        "key_rightshift": [B("KEY_RIGHTSHIFT")],  # 54
        "key_rightalt": [B("KEY_RIGHTALT")],  # 100
        "key_leftmeta": [B("KEY_LEFTMETA")],  # 125
        "key_rightmeta": [B("KEY_RIGHTMETA")],  # 126
        "key_capslock": [B("KEY_CAPSLOCK")],  # 58
        "key_numlock": [B("KEY_NUMLOCK")],  # 69
        "key_scrolllock": [B("KEY_SCROLLLOCK")],  # 70
        "key_sysrq": [B("KEY_SYSRQ")],  # 99
        "key_minus": [B("KEY_MINUS")],  # 12
        "key_equal": [B("KEY_EQUAL")],  # 13
        "key_backspace": [B("KEY_BACKSPACE")],  # 14
        "key_tab": [B("KEY_TAB")],  # 15
        "key_leftbrace": [B("KEY_LEFTBRACE")],  # 26
        "key_rightbrace": [B("KEY_RIGHTBRACE")],  # 27
        "key_space": [B("KEY_SPACE")],  # 57
        "key_up": [B("KEY_UP")],  # 103
        "key_left": [B("KEY_LEFT")],  # 105
        "key_right": [B("KEY_RIGHT")],  # 106
        "key_down": [B("KEY_DOWN")],  # 108
        "key_home": [B("KEY_HOME")],  # 102
        "key_end": [B("KEY_END")],  # 107
        "key_pageup": [B("KEY_PAGEUP")],  # 104
        "key_pagedown": [B("KEY_PAGEDOWN")],  # 109
        "key_insert": [B("KEY_INSERT")],  # 110
        "key_delete": [B("KEY_DELETE")],  # 111
        "key_semicolon": [B("KEY_SEMICOLON")],  # 39
        "key_apostrophe": [B("KEY_APOSTROPHE")],  # 40
        "key_grave": [B("KEY_GRAVE")],  # 41
        "key_backslash": [B("KEY_BACKSLASH")],  # 43
        "key_comma": [B("KEY_COMMA")],  # 51
        "key_dot": [B("KEY_DOT")],  # 52
        "key_slash": [B("KEY_SLASH")],  # 53
        "key_102nd": [B("KEY_102ND")],  # 86
        "key_ro": [B("KEY_RO")],  # 89
        "key_power": [B("KEY_POWER")],  # 116
        "key_compose": [B("KEY_COMPOSE")],  # 127
        "key_stop": [B("KEY_STOP")],  # 128
        "key_again": [B("KEY_AGAIN")],  # 129
        "key_props": [B("KEY_PROPS")],  # 130
        "key_undo": [B("KEY_UNDO")],  # 131
        "key_front": [B("KEY_FRONT")],  # 132
        "key_copy": [B("KEY_COPY")],  # 133
        "key_open": [B("KEY_OPEN")],  # 134
        "key_paste": [B("KEY_PASTE")],  # 135
        "key_cut": [B("KEY_CUT")],  # 137
        "key_find": [B("KEY_FIND")],  # 136
        "key_help": [B("KEY_HELP")],  # 138
        "key_calc": [B("KEY_CALC")],  # 140
        "key_sleep": [B("KEY_SLEEP")],  # 142
        "key_www": [B("KEY_WWW")],  # 150
        "key_screenlock": [B("KEY_SCREENLOCK")],  # 152
        "key_back": [B("KEY_BACK")],  # 158
        "key_refresh": [B("KEY_REFRESH")],  # 173
        "key_edit": [B("KEY_EDIT")],  # 176
        "key_scrollup": [B("KEY_SCROLLUP")],  # 177
        "key_scrolldown": [B("KEY_SCROLLDOWN")],  # 178
        "key_1": [B("KEY_1")],  # 2
        "key_2": [B("KEY_2")],  # 3
        "key_3": [B("KEY_3")],  # 4
        "key_4": [B("KEY_4")],  # 5
        "key_5": [B("KEY_5")],  # 6
        "key_6": [B("KEY_6")],  # 7
        "key_7": [B("KEY_7")],  # 8
        "key_8": [B("KEY_8")],  # 9
        "key_9": [B("KEY_9")],  # 10
        "key_0": [B("KEY_0")],  # 11
        "key_a": [B("KEY_A")],  # 30
        "key_b": [B("KEY_B")],  # 48
        "key_c": [B("KEY_C")],  # 46
        "key_d": [B("KEY_D")],  # 32
        "key_e": [B("KEY_E")],  # 18
        "key_f": [B("KEY_F")],  # 33
        "key_g": [B("KEY_G")],  # 34
        "key_h": [B("KEY_H")],  # 35
        "key_i": [B("KEY_I")],  # 23
        "key_j": [B("KEY_J")],  # 36
        "key_k": [B("KEY_K")],  # 37
        "key_l": [B("KEY_L")],  # 38
        "key_m": [B("KEY_M")],  # 50
        "key_n": [B("KEY_N")],  # 49
        "key_o": [B("KEY_O")],  # 24
        "key_p": [B("KEY_P")],  # 25
        "key_q": [B("KEY_Q")],  # 16
        "key_r": [B("KEY_R")],  # 19
        "key_s": [B("KEY_S")],  # 31
        "key_t": [B("KEY_T")],  # 20
        "key_u": [B("KEY_U")],  # 22
        "key_v": [B("KEY_V")],  # 47
        "key_w": [B("KEY_W")],  # 17
        "key_x": [B("KEY_X")],  # 45
        "key_y": [B("KEY_Y")],  # 21
        "key_z": [B("KEY_Z")],  # 44
        "key_kpasterisk": [B("KEY_KPASTERISK")],  # 55
        "key_kpminus": [B("KEY_KPMINUS")],  # 74
        "key_kpplus": [B("KEY_KPPLUS")],  # 78
        "key_kpdot": [B("KEY_KPDOT")],  # 83
        "key_kpjpcomma": [B("KEY_KPJPCOMMA")],  # 95
        "key_kpenter": [B("KEY_KPENTER")],  # 96
        "key_kpslash": [B("KEY_KPSLASH")],  # 98
        "key_kpequal": [B("KEY_KPEQUAL")],  # 117
        "key_kpcomma": [B("KEY_KPCOMMA")],  # 121
        "key_kpleftparen": [B("KEY_KPLEFTPAREN")],  # 179
        "key_kprightparen": [B("KEY_KPRIGHTPAREN")],  # 180
        "key_kp0": [B("KEY_KP0")],  # 82
        "key_kp1": [B("KEY_KP1")],  # 79
        "key_kp2": [B("KEY_KP2")],  # 80
        "key_kp3": [B("KEY_KP3")],  # 81
        "key_kp4": [B("KEY_KP4")],  # 75
        "key_kp5": [B("KEY_KP5")],  # 76
        "key_kp6": [B("KEY_KP6")],  # 77
        "key_kp7": [B("KEY_KP7")],  # 71
        "key_kp8": [B("KEY_KP8")],  # 72
        "key_kp9": [B("KEY_KP9")],  # 73
        "key_f1": [B("KEY_F1")],  # 59
        "key_f2": [B("KEY_F2")],  # 60
        "key_f3": [B("KEY_F3")],  # 61
        "key_f4": [B("KEY_F4")],  # 62
        "key_f5": [B("KEY_F5")],  # 63
        "key_f6": [B("KEY_F6")],  # 64
        "key_f7": [B("KEY_F7")],  # 65
        "key_f8": [B("KEY_F8")],  # 66
        "key_f9": [B("KEY_F9")],  # 67
        "key_f11": [B("KEY_F11")],  # 87
        "key_f12": [B("KEY_F12")],  # 88
        "key_f10": [B("KEY_F10")],  # 68
        "key_f13": [B("KEY_F13")],  # 183
        "key_f14": [B("KEY_F14")],  # 184
        "key_f15": [B("KEY_F15")],  # 185
        "key_f16": [B("KEY_F16")],  # 186
        "key_f17": [B("KEY_F17")],  # 187
        "key_f18": [B("KEY_F18")],  # 188
        "key_f19": [B("KEY_F19")],  # 189
        "key_f20": [B("KEY_F20")],  # 190
        "key_f21": [B("KEY_F21")],  # 191
        "key_f22": [B("KEY_F22")],  # 192
        "key_f23": [B("KEY_F23")],  # 193
        "key_f24": [B("KEY_F24")],  # 194
        "key_playpause": [B("KEY_PLAYPAUSE")],  # 164
        "key_pause": [B("KEY_PAUSE")],  # 119
        "key_mute": [B("KEY_MUTE")],  # 113
        "key_stopcd": [B("KEY_STOPCD")],  # 166
        "key_forward": [B("KEY_FORWARD")],  # 159
        "key_ejectcd": [B("KEY_EJECTCD")],  # 161
        "key_nextsong": [B("KEY_NEXTSONG")],  # 163
        "key_previoussong": [B("KEY_PREVIOUSSONG")],  # 165
        "key_volumedown": [B("KEY_VOLUMEDOWN")],  # 114
        "key_volumeup": [B("KEY_VOLUMEUP")],  # 115
        "key_katakana": [B("KEY_KATAKANA")],  # 90
        "key_hiragana": [B("KEY_HIRAGANA")],  # 91
        "key_henkan": [B("KEY_HENKAN")],  # 92
        "key_katakanahiragana": [B("KEY_KATAKANAHIRAGANA")],  # 93
        "key_muhenkan": [B("KEY_MUHENKAN")],  # 94
        "key_zenkakuhankaku": [B("KEY_ZENKAKUHANKAKU")],  # 85
        "key_hanguel": [B("KEY_HANGUEL")],  # 122
        "key_hanja": [B("KEY_HANJA")],  # 123
        "key_yen": [B("KEY_YEN")],  # 124
        "key_unknown": [B("KEY_UNKNOWN")],  # 240,
    }
)

__all__ = ["GenericGamepadEvdev", "XBOX_BUTTON_MAP", "XBOX_AXIS_MAP", "B", "to_map"]
